// --------------------------------------------------------------------
// --------------------------------------------------------------------
//
// Major Functions: Video output from Stack machine 
//		VGA state is in SRAM 
//		Three stack cpus share SRAM with VGA via an interleaved switch
// --------------------------------------------------------------------
//
// Revision History :
// --------------------------------------------------------------------
// Bruce R Land, Cornell University, Nov 2011 -- multiproc interface
// Bruce R Land, Cornell University, Nov 2010 -- fast cpu and VGA
// Improved top module written by Adam Shapiro Oct 2009
// --------------------------------------------------------------------

`include "scan_events.h"

module DE2_TOP (
    // Clock Input
    input         CLOCK_27,    // 27 MHz
    input         CLOCK_50,    // 50 MHz
    input         EXT_CLOCK,   // External Clock
    // Push Button
    input  [3:0]  KEY,         // Pushbutton[3:0]
    // DPDT Switch
    input  [17:0] SW,          // Toggle Switch[17:0]
    // 7-SEG Display
    output [6:0]  HEX0,        // Seven Segment Digit 0
    output [6:0]  HEX1,        // Seven Segment Digit 1
    output [6:0]  HEX2,        // Seven Segment Digit 2
    output [6:0]  HEX3,        // Seven Segment Digit 3
    output [6:0]  HEX4,        // Seven Segment Digit 4
    output [6:0]  HEX5,        // Seven Segment Digit 5
    output [6:0]  HEX6,        // Seven Segment Digit 6
    output [6:0]  HEX7,        // Seven Segment Digit 7
    // LED
    output [8:0]  LEDG,        // LED Green[8:0]
    output [17:0] LEDR,        // LED Red[17:0]
    // UART
    output        UART_TXD,    // UART Transmitter
    input         UART_RXD,    // UART Receiver
    // IRDA
    output        IRDA_TXD,    // IRDA Transmitter
    input         IRDA_RXD,    // IRDA Receiver
    // SDRAM Interface
    inout  [15:0] DRAM_DQ,     // SDRAM Data bus 16 Bits
    output [11:0] DRAM_ADDR,   // SDRAM Address bus 12 Bits
    output        DRAM_LDQM,   // SDRAM Low-byte Data Mask 
    output        DRAM_UDQM,   // SDRAM High-byte Data Mask
    output        DRAM_WE_N,   // SDRAM Write Enable
    output        DRAM_CAS_N,  // SDRAM Column Address Strobe
    output        DRAM_RAS_N,  // SDRAM Row Address Strobe
    output        DRAM_CS_N,   // SDRAM Chip Select
    output        DRAM_BA_0,   // SDRAM Bank Address 0
    output        DRAM_BA_1,   // SDRAM Bank Address 0
    output        DRAM_CLK,    // SDRAM Clock
    output        DRAM_CKE,    // SDRAM Clock Enable
    // Flash Interface
    inout  [7:0]  FL_DQ,       // FLASH Data bus 8 Bits
    output [21:0] FL_ADDR,     // FLASH Address bus 22 Bits
    output        FL_WE_N,     // FLASH Write Enable
    output        FL_RST_N,    // FLASH Reset
    output        FL_OE_N,     // FLASH Output Enable
    output        FL_CE_N,     // FLASH Chip Enable
    // SRAM Interface
    inout  [15:0] SRAM_DQ,     // SRAM Data bus 16 Bits
    output [17:0] SRAM_ADDR,   // SRAM Address bus 18 Bits
    output        SRAM_UB_N,   // SRAM High-byte Data Mask 
    output        SRAM_LB_N,   // SRAM Low-byte Data Mask 
    output        SRAM_WE_N,   // SRAM Write Enable
    output        SRAM_CE_N,   // SRAM Chip Enable
    output        SRAM_OE_N,   // SRAM Output Enable
    // ISP1362 Interface
    inout  [15:0] OTG_DATA,    // ISP1362 Data bus 16 Bits
    output [1:0]  OTG_ADDR,    // ISP1362 Address 2 Bits
    output        OTG_CS_N,    // ISP1362 Chip Select
    output        OTG_RD_N,    // ISP1362 Write
    output        OTG_WR_N,    // ISP1362 Read
    output        OTG_RST_N,   // ISP1362 Reset
    output        OTG_FSPEED,  // USB Full Speed, 0 = Enable, Z = Disable
    output        OTG_LSPEED,  // USB Low Speed,  0 = Enable, Z = Disable
    input         OTG_INT0,    // ISP1362 Interrupt 0
    input         OTG_INT1,    // ISP1362 Interrupt 1
    input         OTG_DREQ0,   // ISP1362 DMA Request 0
    input         OTG_DREQ1,   // ISP1362 DMA Request 1
    output        OTG_DACK0_N, // ISP1362 DMA Acknowledge 0
    output        OTG_DACK1_N, // ISP1362 DMA Acknowledge 1
    // LCD Module 16X2
    inout  [7:0]  LCD_DATA,    // LCD Data bus 8 bits
    output        LCD_ON,      // LCD Power ON/OFF
    output        LCD_BLON,    // LCD Back Light ON/OFF
    output        LCD_RW,      // LCD Read/Write Select, 0 = Write, 1 = Read
    output        LCD_EN,      // LCD Enable
    output        LCD_RS,      // LCD Command/Data Select, 0 = Command, 1 = Data
    // SD Card Interface
    inout         SD_DAT,      // SD Card Data
    inout         SD_DAT3,     // SD Card Data 3
    inout         SD_CMD,      // SD Card Command Signal
    output        SD_CLK,      // SD Card Clock
    // I2C
    inout         I2C_SDAT,    // I2C Data
    output        I2C_SCLK,    // I2C Clock
    // PS2
    input         PS2_DAT,     // PS2 Data
    input         PS2_CLK,     // PS2 Clock
    // USB JTAG link
    input         TDI,         // CPLD -> FPGA (data in)
    input         TCK,         // CPLD -> FPGA (clk)
    input         TCS,         // CPLD -> FPGA (CS)
    output        TDO,         // FPGA -> CPLD (data out)
    // VGA
    output        VGA_CLK,     // VGA Clock
    output        VGA_HS,      // VGA H_SYNC
    output        VGA_VS,      // VGA V_SYNC
    output        VGA_BLANK,   // VGA BLANK
    output        VGA_SYNC,    // VGA SYNC
    output [9:0]  VGA_R,       // VGA Red[9:0]
    output [9:0]  VGA_G,       // VGA Green[9:0]
    output [9:0]  VGA_B,       // VGA Blue[9:0]
    // Ethernet Interface
    inout  [15:0] ENET_DATA,   // DM9000A DATA bus 16Bits
    output        ENET_CMD,    // DM9000A Command/Data Select, 0 = Command, 1 = Data
    output        ENET_CS_N,   // DM9000A Chip Select
    output        ENET_WR_N,   // DM9000A Write
    output        ENET_RD_N,   // DM9000A Read
    output        ENET_RST_N,  // DM9000A Reset
    input         ENET_INT,    // DM9000A Interrupt
    output        ENET_CLK,    // DM9000A Clock 25 MHz
    // Audio CODEC
    inout         AUD_ADCLRCK, // Audio CODEC ADC LR Clock
    input         AUD_ADCDAT,  // Audio CODEC ADC Data
    inout         AUD_DACLRCK, // Audio CODEC DAC LR Clock
    output        AUD_DACDAT,  // Audio CODEC DAC Data
    inout         AUD_BCLK,    // Audio CODEC Bit-Stream Clock
    output        AUD_XCK,     // Audio CODEC Chip Clock
    // TV Decoder
    input  [7:0]  TD_DATA,     // TV Decoder Data bus 8 bits
    input         TD_HS,       // TV Decoder H_SYNC
    input         TD_VS,       // TV Decoder V_SYNC
    output        TD_RESET,    // TV Decoder Reset
    // GPIO
    inout  [35:0] GPIO_0,      // GPIO Connection 0
    inout  [35:0] GPIO_1       // GPIO Connection 1
);
    
   //Turn off hex displays.
   assign HEX0 = 7'h7F;
   assign HEX1 = 7'h7F;
   assign HEX2 = 7'h7F;
   assign HEX3 = 7'h7F;
   assign HEX4 = 7'h7F;
   assign HEX5 = 7'h7F;
   assign HEX6 = 7'h7F;
   assign HEX7 = 7'h7F;
   assign LEDG = is_caps_lock_on;
   
   //Set all GPIO to tri-state.
   //assign GPIO_0 = 36'hzzzzzzzzz;
   assign GPIO_1 = 36'hzzzzzzzzz;

   //Disable audio codec.
   //assign AUD_DACDAT = 1'b0;
   //assign AUD_XCK    = 1'b0;

   //Disable DRAM.
   assign DRAM_ADDR  = 12'h0;
   assign DRAM_BA_0  = 1'b0;
   assign DRAM_BA_1  = 1'b0;
   assign DRAM_CAS_N = 1'b1;
   assign DRAM_CKE   = 1'b0;
   assign DRAM_CLK   = 1'b0;
   assign DRAM_CS_N  = 1'b1;
   assign DRAM_DQ    = 16'hzzzz;
   assign DRAM_LDQM  = 1'b0;
   assign DRAM_RAS_N = 1'b1;
   assign DRAM_UDQM  = 1'b0;
   assign DRAM_WE_N  = 1'b1;

   //Disable Ethernet.
   assign ENET_CLK   = 1'b0;
   assign ENET_CS_N  = 1'b1;
   assign ENET_CMD   = 1'b0;
   assign ENET_DATA  = 16'hzzzz;
   assign ENET_RD_N  = 1'b1;
   assign ENET_RST_N = 1'b1;
   assign ENET_WR_N  = 1'b1;

   //Disable flash.
   assign FL_ADDR  = 22'h0;
   assign FL_CE_N  = 1'b1;
   assign FL_DQ    = 8'hzz;
   assign FL_OE_N  = 1'b1;
   assign FL_RST_N = 1'b1;
   assign FL_WE_N  = 1'b1;

   //Disable LCD.
   assign LCD_BLON = 1'b0;
   assign LCD_DATA = 8'hzz;
   assign LCD_EN   = 1'b0;
   assign LCD_ON   = 1'b0;
   assign LCD_RS   = 1'b0;
   assign LCD_RW   = 1'b0;

   //Disable OTG.
   assign OTG_ADDR    = 2'h0;
   assign OTG_CS_N    = 1'b1;
   assign OTG_DACK0_N = 1'b1;
   assign OTG_DACK1_N = 1'b1;
   assign OTG_FSPEED  = 1'b1;
   assign OTG_DATA    = 16'hzzzz;
   assign OTG_LSPEED  = 1'b1;
   assign OTG_RD_N    = 1'b1;
   assign OTG_RST_N   = 1'b1;
   assign OTG_WR_N    = 1'b1;

   //Disable all other peripherals.
   //assign I2C_SCLK = 1'b0;
   assign IRDA_TXD = 1'b0;
   //assign TD_RESET = 1'b0;
   assign TDO = 1'b0;
   assign UART_TXD = 1'b0;

// === VGA timing and interface =================
wire	VGA_CTRL_CLK;
wire	AUD_CTRL_CLK;
wire	FAST_CLK;
wire	DLY_RST;

assign	TD_RESET	=	1'b1;	//	Allow 27 MHz
assign	AUD_ADCLRCK	=	AUD_DACLRCK;
assign	AUD_XCK		=	AUD_CTRL_CLK;

wire reset ;
assign reset = SW[0];

Reset_Delay			r0	( .iCLK(CLOCK_27), .oRESET(DLY_RST) );

wire sram_clk;

// sram_clk and FAST_CLK are 50.4 MHz -- VGA_CLK is 25.2
VGA_Audio_PLL 		p1	(
    .areset(~DLY_RST),
    .inclk0(CLOCK_27),
    .c0(sram_clk),
    .c1(FAST_CLK),
    .c2(VGA_CLK) ); //.areset(~DLY_RST)

VGA_Controller		u1	(	//	Host Side
    .iCursor_RGB_EN(4'b0111),
    .oAddress(mVGA_ADDR),
    .oCoord_X(Coord_X),
    .oCoord_Y(Coord_Y),
    .iRed(mVGA_R),
    .iGreen(mVGA_G),
    .iBlue(mVGA_B),
    //	VGA Side
    .oVGA_R(VGA_R),
    .oVGA_G(VGA_G),
    .oVGA_B(VGA_B),
    .oVGA_H_SYNC(VGA_HS),
    .oVGA_V_SYNC(VGA_VS),
    .oVGA_SYNC(VGA_SYNC),
    .oVGA_BLANK(VGA_BLANK),
    //	Control Signal
    .iCLK(VGA_CLK),         //VGA_CTRL_CLK
    .iRST_N(DLY_RST)	);  //.iRST_N(DLY_RST)

wire [9:0]	mVGA_R;				//memory output to VGA
wire [9:0]	mVGA_G;
wire [9:0]	mVGA_B;
wire [18:0]	mVGA_ADDR;			//video memory address
wire [9:0]  Coord_X, Coord_Y;	//display coods



//use LEDR for debugging
reg [13:0] test_led;
reg        success_led;
reg        error_led;
reg        on_led;
assign LEDR = {on_led, success_led, error_led, toggle, test_led};


//=============== SD CARD ===============//

//http://www.theoengel.nl/P8/index.html

`define SD_CMD_READ 2'b01
`define SD_CMD_WRITE 2'b10
`define SD_CMD_NONE 2'b00

//makes it easier to change clk for all SD-related blocks
wire sdc_clk;
assign sdc_clk = FAST_CLK;

// maintain state machine for SD card operations
typedef enum {  // enums available in SystemVerilog
    init, init2, init3, init4, init5, init6,
    write, write2, write3, write4, write5, write6, write7, write8, write9, write10, write11,
	read, read2, read3, read4, read5, read6, read7, read8, read9, read10, read11, read12, read13,
    idle, error, error2
} sd_state_e;

// current state of SD card init/read/write
sd_state_e sd_state;

// inputs to SD_Card module
reg  [1:0]  sd_cmd;         // Read (01), Write (10), or None (00) (default)
wire [7:0]  sd_write_data;  // output data byte to sdcard
wire [15:0] sd_sector_num;  // sector number from 0 to 50,000
wire [31:0] sd_addr;        // Byte Address

assign sd_write_data = cpu1_out7;
assign sd_sector_num = cpu1_out6;
assign sd_addr = {7'b0, sd_sector_num, 9'b0};

// outputs from SD_Card module
wire        sd_cmd_ack;     // signal to host that read/write command is accepted
wire        sd_rw_ready;    // signal to host that read/write command is processed
wire        sd_break;       // signal an input break (byte available from sdcard)
wire [7:0]  sd_read_data;   // data byte from sdcard (if oBreak == 1)
wire        sd_ack_output;  // signal to inform the host that output byte is being sent to sd card
wire        sd_operational; // sd card is in init state/ready
wire [7:0]  sd_response;    // sd card response byte of a write command
reg  [1:0]  sd_data_valid;  // 01 if data from read is valid, 10 if data from write valid, 00 if none valid
reg  [7:0]  sd_read_data_2cpu;
reg  [7:0]  sd_read_data_tmp;
reg         break_flag;

reg         sw1_prev, sw2_prev;

SD_Card sd(
    .iRST_n(        reset),
    .iCLK(          sdc_clk),
    .iCmd(          sd_cmd),
    .iScAd(         sd_addr),
    .iWsD_DATA(     sd_write_data),
    
    .oCmd_Ack(      sd_cmd_ack),
    .oOperational(  sd_operational),
    .oRW_Ready(     sd_rw_ready),
    .oBreak(        sd_break),
    .oRsD_DATA(     sd_read_data),
    .Response(      sd_response),
    .oAckOutput(    sd_ack_output),
    
    .SD_CMD(        SD_CMD),
    .SD_DAT(        SD_DAT),
    .SD_DAT3(       SD_DAT3),
    .SD_CLK(        SD_CLK)
);

always @(posedge sdc_clk) begin
    if (~reset) begin
        sd_cmd <= `SD_CMD_NONE;
        sd_state <= init;
        
        on_led <= 0;
        test_led <= 0;
        success_led <= 0;
        error_led <= 0;
        
        sd_data_valid <= 2'b00;
        break_flag <= 0;
        
        sw1_prev <= SW[1];
        sw2_prev <= SW[2];
    end
    else begin
        on_led <= 1;
        case(sd_state)
            init:
            begin
                if (sd_operational == 1'b1) begin
                    sd_state <= idle;
                end
            end
            write:
            begin
                sd_cmd <= `SD_CMD_WRITE;
                sd_data_valid <= 2'b00;
                
                test_led[7:0] <= sd_write_data;
                test_led[11:8] <= sd_sector_num[3:0];
                
                if (sd_rw_ready == 1'b1) begin
                    sd_cmd <= `SD_CMD_NONE;
                    sd_state <= write2;
                end
            end
            write2:
            begin
                // check the response of the write
                // all 0's means no errors
                if (sd_response == 8'h00) begin
                    success_led <= 1;
                    
                    sd_state <= idle;
                    sd_data_valid <= `SD_CMD_WRITE; //signal cpu that write is done
                end
                else begin
                    test_led <= sd_response;
                    sd_state <= error;
                end
            end
            read:
            begin
                sd_cmd <= `SD_CMD_READ;
                sd_data_valid <= 0;
                
                if ((sd_break == 1'b1) && (break_flag == 1'b0)) begin
                    // data available
                    sd_read_data_tmp <= sd_read_data;
                    break_flag <= 1'b1;
                end
                
                if (sd_rw_ready == 1'b1) begin
                    sd_read_data_2cpu <= sd_read_data_tmp;
                    sd_data_valid <= `SD_CMD_READ; //signal cpu that read is done
                    
                    success_led <= 1;
                    test_led[7:0] <= sd_read_data_tmp;
                    test_led[11:8] <= sd_sector_num[3:0];
                    
                    sd_cmd <= `SD_CMD_NONE;
                    sd_state <= idle;
                end
            end
            idle:
            begin
                if (sd_sector_num == -18'b1) begin
                    //no byte is being requested by CPU
                    sd_state <= idle;
                end
                //wait here for a new command
                else if (cpu1_out5 == `SD_CMD_WRITE) begin //start write
                    sd_state <= write;
                end
                else if (cpu1_out5 == `SD_CMD_READ) begin //start read
                    break_flag <= 0;
                    sd_state <= read;
                end
                else if (cpu1_out5 == `SD_CMD_NONE) begin
                    sd_data_valid <= 0; //read byte no longer valid
                end
            end
            error:
            begin
                sd_cmd <= `SD_CMD_NONE;
                error_led <= 1;
                success_led <= 0;
                sd_data_valid <= 2'b00;
            end
        endcase
    end
end


//=============== KEYBOARD ===============//
//Based on Skyler Schneider's PS2 keyboard controller (ECE 5760, FA 2010)

reg        toggle;
reg  [7:0] key_pressed;
reg        is_shift_pressed;
reg        is_caps_lock_on;
wire [7:0] k_lut_out;

//makes it easier to change clk for all keyboard-related blocks
wire keyboard_clk;
assign keyboard_clk = FAST_CLK;

wire        wKeyboardEventReady;
wire [7:0]  wKeyboardEventType;
Keyboard_Controller kc0 (
    .iPS2_CLK    (PS2_CLK),
    .iPS2_DAT    (PS2_DAT),
    .iCLK        (keyboard_clk),
    .iRESET      (~reset),
    .oEventReady (wKeyboardEventReady),
    .oEventType  (wKeyboardEventType)
);


keyLUT k_lut(
    .scancode(wKeyboardEventType), //in
    .shift(is_shift_pressed),      //in
    .caps_lock(is_caps_lock_on),   //in
    .ascii(k_lut_out)              //out
);

always @(posedge keyboard_clk) begin
    if (~reset) begin
        toggle <= 1'b0;
		key_pressed <= 8'b0;
        is_shift_pressed <= 0;
        is_caps_lock_on <= 0;
    end

    else if ((wKeyboardEventType == `SC_MAKE_LF_SHIFT) || (wKeyboardEventType == `SC_MAKE_RT_SHIFT)) begin
        is_shift_pressed <= 1;
    end
    else if ((wKeyboardEventType == `SC_BREAK_LF_SHIFT) || (wKeyboardEventType == `SC_BREAK_RT_SHIFT)) begin
        is_shift_pressed <= 0;
    end
    else if (wKeyboardEventType == `SC_MAKE_CAPS_LOCK) begin
        is_caps_lock_on <= ~is_caps_lock_on;
    end
    else if (wKeyboardEventType >= `SC_MAKE_TILDE && wKeyboardEventType <= `SC_MAKE_SPACE) begin
        key_pressed <= k_lut_out;
        toggle <= ~toggle;
    end
end


// =============== SRAM interface ===============
// 18 bits of adddress:
// 9 bits of Y 
// 10 bits of X (excludes low bit which drives LB/UB)
// reset high is normal operation
reg [17:0] sram_addr_reg ;
reg [15:0] sram_data_reg ;
reg sram_we, sram_ub, sram_lb ;

assign SRAM_ADDR = sram_addr_reg ;
assign SRAM_CE_N = 1'b0; // always enabled
assign SRAM_OE_N = 1'b0; // always enabled
assign SRAM_LB_N = sram_lb;
assign SRAM_UB_N = sram_ub;
assign SRAM_WE_N = sram_we;
assign SRAM_DQ   = sram_we? 16'hzzzz : sram_data_reg ;

reg  Coord_x_0 ; // low bit of vga x

// =============== memory switch ===============
// now allow VGA and cpu to access SRAM
// VGA gets SRAM on every other memory clock cycle
// (once per VGA cycle).
// cpu gets SRAM on every other cycle.
// SRAM clock is same frequency as cpu clock (50.4 MHZ)
// but is phase shifted +90 degrees in the PLL.

wire [17:0] cpu1_out0, cpu1_out1, cpu1_out2, cpu1_out3, cpu1_out4, cpu1_out5, cpu1_out6, cpu1_out7 ;

// data from SRAM back to cpus
reg [15:0] cpu1_sram_data, vga_sram_data ;
// feedback to cpu telling cpu that memory is theirs
reg cpu1_access;
// internal time counter to account for slow software
reg [3:0] cpu1_wait;

parameter wait_time = 4'd8 ; //4'd10 the wait time for a cpu to clear it's request flag

// sram control word:
// bit 0 low => we
// bit 1 low => write low byte
// bit 1 high=> write high byte
// bit 2 high => request sram access
// bit 3 high => full word read/write
// bit 4 high => mutex test and set
// bit 5 high => mutex read
// bit 6 high => mutex clear
// ==== Memory protocol ======
// Processor outputs address (and data on write).
// Processor requests memory (read, write, mutex operation)
// Processor waits for memory to ACK access
// On read, processsor reads memory. 
// Processor clears request 
// --(memory delays a few cycles to allow proecssor to clear request)

always @(posedge sram_clk) //sram_clk
begin
	
	if (~reset)
	begin
		sram_addr_reg <= {Coord_Y[8:0], Coord_X[9:1]} ;
		sram_data_reg  <= 16'b0 ;
		sram_lb <= 1'b0 ; 
		sram_ub <=  1'b0 ;
		sram_we <= 1'b0 ;
		// tell all cpus that they don't have access
		cpu1_access <= 1'b0 ;
		// init cpu wait functions
		cpu1_wait <= 1'b0 ;
	end
		
	// on VGA_CLK allow VGA controller access
	// set up read, then finish read on next cycle
	else if (VGA_CLK) //(clk_phase) note that sram_clk is phase-locked and twice as fast
	begin
		sram_addr_reg <= {Coord_Y[8:0], Coord_X[9:1]} ;
		sram_lb <= 1'b0 ; 
		sram_ub <=  1'b0 ; 
		sram_we <=  1'b1 ;	
	end

	// on ~VGA_CLK allow CPU access
	// set up read, then finish read on next cycle	
	// sram_cntl word:
	// bit 0 low => we
	// bit 1 low => write low byte
	// bit 1 high=> write high byte
	// bit 2 high => request sram access
	// bit 3 high => full word read/write
	// bit 4 high => mutex test and set
	// bit 5 high => mutex read
	// bit 6 high => mutex clear
	else  // if (~VGA_CLK)
	begin
		if (cpu1_out2[2] && cpu1_wait==0 ) 
		begin
			sram_addr_reg <= cpu1_out0 ;
			sram_data_reg <= cpu1_out1 ; 
			sram_lb <= (cpu1_out2[3] | cpu1_out2[0])? 0 : cpu1_out2[1] ; 
			sram_ub <= (cpu1_out2[3] | cpu1_out2[0])? 0 : ~cpu1_out2[1] ; 
			sram_we <= cpu1_out2[0] ;
			cpu1_access <= 1'b1 ;
			cpu1_wait <= 1'b1 ;
		end
	end
	
	// read/write data on opposite cycle from 
	// previous setup steps above
	// get data 
	if (VGA_CLK && cpu1_wait==1)
	begin
		cpu1_sram_data <= SRAM_DQ ;
	end
	
	if (~VGA_CLK)
	begin
		vga_sram_data <= SRAM_DQ ;
		Coord_x_0 <= (Coord_X[0]);
	end	
	
	// memory time outs for each cpu
	if (cpu1_wait>0)
	begin
		if (cpu1_wait < wait_time) // 
			cpu1_wait <= cpu1_wait + 4'd1 ;
		else
		begin
			cpu1_wait <= 4'b0 ;
			cpu1_access <= 1'b0 ;
		end
	end

end // @(posedge sram_clk)

// =============== color map ===============
// remap color index for bigger dynamic range
wire [9:0] r, g, b ;
assign mVGA_R = r ;
assign mVGA_G = g ;
assign mVGA_B = b ;
wire [7:0] index ;
assign index = (Coord_x_0)? vga_sram_data[15:8] : vga_sram_data[7:0] ;

// define the vga colormap
colormap cmap(r, g, b, index);

// =============== font rom ===============
wire [15:0] cpu1_in4;
char_rom rom1(FAST_CLK, cpu1_out4, cpu1_in4);

// =============== Instantiate minicpu ===============
minicpu cpu1(.clk(FAST_CLK), .reset(reset),  .run(1'b1),
	.in0(sd_read_data_2cpu),  // data read from SD card
    .in1(sd_data_valid), // whether data from SD card is valid
	.in2({2'b00,cpu1_sram_data}),
	.in3(VGA_VS),     // VGA vertical sync
	.in4(cpu1_in4),   // char rom data
	.in5(key_pressed),// the key on the keyboard that was pressed
	.in6(toggle),     // toggles each time there is a new key pressed
    .in7({cpu1_access}), // can remove this if more in ports are needed
	.out0(cpu1_out0), // sram address
	.out1(cpu1_out1), // sram data
	.out2(cpu1_out2), // sram control
    .out3(),
	.out4(cpu1_out4), // char rom address
    .out5(cpu1_out5), // SDC command (read 01, write 10, none 00)
    .out6(cpu1_out6), // byte/sector on SDC to be read (0-50,000 or -1 if none)
    .out7(cpu1_out7)
);

endmodule //top module



/////////////////////////////
//    VGA CHARACTER ROM    //
/////////////////////////////

module char_rom(clk, addr, q) /* synthesis romstyle = "M4K" */ ;
 parameter DWIDTH=8,AWIDTH=11,WORDS=2048;

 input clk ;
 input [AWIDTH-1:0] addr;
 output [DWIDTH-1:0] q;
 reg [DWIDTH-1:0] q;
 // IF YOUR PROGRAM IS IN a mif file use:
 // reg [DWIDTH-1:0] mem [WORDS-1:0] /* synthesis ram_init_file = "TestPgm0.mif" */ ;
 // see http://www.altera.com/literature/hb/qts/qts_qii51008.pdf EXAMPLE 9-78
 
 reg [DWIDTH-1:0] mem [WORDS-1:0]  /* synthesis ram_init_file = "font_rom.mif" */ ;
 
// initial // Read the memory contents in the file
//dual_port_rom_init.txt.
//begin
//$readmemb("dual_port_rom_init.txt", rom);
//end

 always @(posedge clk)
   begin
     q <= mem[addr];
   end

endmodule

/////////////////////
//    COLOR MAP    //
/////////////////////

module colormap(r,g,b, index);
// index bits r=[7:5], g=[4:2], b=[1:0]
input [7:0] index;
output reg [9:0] r, g, b ;
always @(*)
begin
if (index[7:6]==2'b11)
begin
	// red map
	case(index[5:4])
		0: r=10'h0;
		1: r=10'h100;
		2: r=10'h200;
		3: r=10'h3ff;
	endcase 
	// green map
	case(index[3:2])
		0: g=10'h0;
		1: g=10'h100;
		2: g=10'h200;
		3: g=10'h3ff;
	endcase 
	// blue map
	case(index[1:0])
		0: b=10'h0;
		1: b=10'h100;
		2: b=10'h200;
		3: b=10'h3ff;
	endcase
end
	
else if (index[7:6]==2'b10) 
begin
	// red map
	case(index[5:4])
		0: r=10'h0;
		1: r=10'h100>>1;
		2: r=10'h200>>1;
		3: r=10'h3ff>>1;
	endcase 
	// green map
	case(index[3:2])
		0: g=10'h0;
		1: g=10'h100>>1;
		2: g=10'h200>>1;
		3: g=10'h3ff>>1;
	endcase 
	// blue map
	case(index[1:0])
		0: b=10'h0;
		1: b=10'h100>>1;
		2: b=10'h200>>1;
		3: b=10'h3ff>>1;
	endcase
end
	
else if (index[7:6]==2'b01) 
begin
	// red map
	case(index[5:4])
		0: r=10'h0;
		1: r=10'h100>>2;
		2: r=10'h200>>2;
		3: r=10'h3ff>>2;
	endcase 
	// green map
	case(index[3:2])
		0: g=10'h0;
		1: g=10'h100>>2;
		2: g=10'h200>>2;
		3: g=10'h3ff>>2;
	endcase 
	// blue map
	case(index[1:0])
		0: b=10'h0;
		1: b=10'h100>>2;
		2: b=10'h200>>2;
		3: b=10'h3ff>>2;
	endcase
end
	
else begin
	// red map
	case(index[5:4])
		0: r=10'h0;
		1: r=10'h100>>3;
		2: r=10'h200>>3;
		3: r=10'h3ff>>3;
	endcase 
	// green map
	case(index[3:2])
		0: g=10'h0;
		1: g=10'h100>>3;
		2: g=10'h200>>3;
		3: g=10'h3ff>>3;
	endcase 
	// blue map
	case(index[1:0])
		0: b=10'h0;
		1: b=10'h100>>3;
		2: b=10'h200>>3;
		3: b=10'h3ff>>3;
	endcase
end
end
endmodule

/////////////////////////////////////////////
//    signed mult of 10.8 format 2'comp    //
/////////////////////////////////////////////
module signed_mult (out, a, b);

	output 		[17:0]	out;
	input 	signed	[17:0] 	a;
	input 	signed	[17:0] 	b;
	
	wire	signed	[17:0]	out;
	wire 	signed	[35:0]	mult_out;

	assign mult_out = a * b;
	//assign out = mult_out[33:17];
	assign out = {mult_out[35], mult_out[24:8]};
endmodule



////////////////////////////////////////////////////////////
// The following CPU design is from                       //
// http://www.cs.hiroshima-u.ac.jp/~nakano/wiki/          //
// and is GPL licensed                                    //
//                                                        //
// with added opcodes, 18-bit data, and modified mult     //
////////////////////////////////////////////////////////////

`define IDLEA 2'b00
`define IDLEB 2'b01
`define EXEC 2'b10

`define  ADD  5'b00000
`define  SUB  5'b00001
`define  MUL  5'b00010
`define  SHL  5'b00011
`define  SHR  5'b00100
`define  BAND 5'b00101
`define  BOR  5'b00110
`define  BXOR 5'b00111
`define  AND  5'b01000
`define  OR   5'b01001
`define  EQ   5'b01010
`define  NE   5'b01011
`define  GE   5'b01100
`define  LE   5'b01101
`define  GT   5'b01110
`define  LT   5'b01111
`define  NEG  5'b10000
`define  BNOT 5'b10001
`define  NOT  5'b10010
`define  A    5'b10011
`define  B    5'b10111
`define  Aa   5'b11011
`define  Ba   5'b11111
`define  ASR  5'b10100 // arithmetic shift right
`define  MFIX 5'b10101 //fixed mult

`define HALT 4'b0000
`define PUSHI 4'b0001
`define PUSH 4'b0010
`define POP  4'b0011
`define JMP  4'b0100
`define JZ  4'b0101
`define JNZ 4'b0110
`define LD  4'b0111
`define ST  4'b1000
`define PPPC 4'b1001
`define IN   4'b1101
`define OUT  4'b1110
`define OP   4'b1111

//////////////////////////////////
module alu(a, b, f, s);

 input [17:0] a, b;
 input [4:0] 	f;
 output [17:0] s;
 reg  [17:0] 	 s;
 wire [17:0] x, y, mout;

 assign x = a + 18'h20000;
 assign y = b + 18'h20000;
  
 always @(a or b or x or y or f or mout)
   case(f)
     `ADD : s = b + a;
     `SUB : s = b - a;
     `MUL : s =  b * a; //mout ; // b * a;
	  `MFIX: s =  mout ;
     `SHL : s = b << a;
     `SHR : s = b >> a;
	  `ASR : s = $signed(b) >>> a;
     `BAND: s = b & a;
     `BOR : s = b | a;
     `BXOR: s = b ^ a;
     `AND : s = b && a;
     `OR  : s = b || a;
     `EQ  : s = b == a;
     `NE  : s = b != a;
     `GE  : s = y >= x;
     `LE  : s = y <= x;
     `GT  : s = y > x;
     `LT  : s = y < x;
     `NEG : s = -a;
     `BNOT : s = ~a;
     `NOT : s = !a;
	  `A   : s = a;
	  `B   : s = b;
	  `Aa  : s = a;
	  `Ba  : s = b;
     default : s = 18'hxxxx;
   endcase

	signed_mult m_10_8(mout, a, b) ;
endmodule

////////////////////////////////////
module statef(clk,reset,run,halt,cs);
   
  input clk, reset, run, halt;
  output [1:0] cs;
  reg [1:0] cs;

  always @(posedge clk or negedge reset)
    if(!reset) cs <= `IDLEA;
    else
      case(cs)
        `IDLEA: if(run) cs <= `IDLEB;
        `IDLEB: cs <= `EXEC;
        `EXEC: if(halt) cs <= `IDLEA;
                else cs <= `EXEC;
		  //`EXEC:  cs <= `EXEC;
        default: cs <= 2'bxx;
      endcase

endmodule


//////////////////////////////////////

module stackm(clk, reset, load, push, pop, pop2, thru, d, dthru, qtop, qnext);
parameter AWIDTH = 5, N = 32;

  input clk, reset, load, push, pop, pop2, thru;
  input [17:0] d, dthru;
  output [17:0] qtop, qnext;
  reg [AWIDTH-1:0] ptop, ptopf, addra, addrb;
  reg [17:0] mem [N-1:0];
  reg [17:0] qtopf;
  wire [17:0] pnextf;
  wire thruwrite;

  always @(push or pop or pop2 or ptop)
  if(push) ptopf = ptop - 1;
  else if(pop) ptopf = ptop + 1;
  else if(pop2) ptopf = ptop + 2;
  else ptopf = ptop;

  always @(posedge clk or negedge reset) 
  if(!reset) ptop <= 0;
  else ptop <= ptopf;

  assign  pnextf = ptopf + 1;
  assign thruwrite = !(push||pop) && thru;

  always @(load or thruwrite or d or dthru)
  if(load) qtopf = d;
  else if(thruwrite) qtopf = dthru;
  else qtopf = 18'hxxxx;

  always @(posedge clk)
  begin
    if(load || thruwrite) mem[ptopf] <= qtopf;
    addra <= ptopf;
  end

  always @(posedge clk)
  begin
    if(push && thru) mem[pnextf] <= dthru;
    addrb <= pnextf;
  end

  assign qtop = (thru ? dthru : mem[addra]);
  assign qnext = mem[addrb];

endmodule


//////////////////////////////////////////////////

module dpram(clk, load1, addr1, addr2, d1, q1, q2);
 parameter DWIDTH=18,AWIDTH=12,WORDS=4096;

 input clk,load1 ; //,load2;
 input [AWIDTH-1:0] addr1,addr2;
 input [DWIDTH-1:0] d1 ; //,d2;
 output [DWIDTH-1:0] q1,q2;
 reg [DWIDTH-1:0] q1,q2;
 reg [DWIDTH-1:0] mem [WORDS-1:0] /* synthesis ram_init_file = "stack_three_vga_sram_access_2013.mif" */ ;

 always @(posedge clk)
   begin
     if(load1) mem[addr1] <= d1;
     q1 <= mem[addr1];
  end

 always @(posedge clk)
   begin
//     if(load2) mem[addr2] <= d2;
     q2 <= mem[addr2];
  end

//  initial begin
//		mem[12'h000]=16'hd000;
//		mem[12'h001]=16'he000;
//		mem[12'h002]=16'h4000;
//  end

endmodule
//////////////////////////////////////

module minicpu(clk, reset, run, 
					in0, in1, in2, in3, in4, in5, in6, in7,
					out0, out1, out2, out3, out4, out5, out6, out7);

 input clk,reset,run;
 input [17:0] in0, in1, in2, in3, in4, in5, in6, in7  ;
 output [17:0] out0, out1, out2, out3, out4, out5, out6, out7 ;
 
 wire [1:0] cs;
 wire [17:0] irout, qtop, dbus ;
 wire [17:0] qnext, ramout, aluout;
 reg [11:0] abus;
 reg halt, jump, pcinc, push, pop, thru, qthru, dbus2qtop;
 reg dbus2ram, dbus2obuf, ir2dbus, qtop2dbus, alu2dbus, ram2dbus, in2dbus;
 reg pop2, ir2abus, qtop2abus, qnext2abus;
 reg poppc, pushpc ; // added by BRL4 to support POPPC/PUSHPC opcode
 reg [11:0] pcout, pcnext;
 reg [17:0] out[7:0] ;
 wire [17:0] in[7:0] ;

 statef statef0(.clk(clk),.reset(reset),.run(run),.halt(halt),.cs(cs));
 stackm stackm0(.clk(clk),.reset(reset),.load(dbus2qtop),.push(push),
	.pop(pop),.pop2(pop2),.thru(qthru),.d(dbus),.dthru(ramout),.qtop(qtop),.qnext(qnext));
 alu alu0(.a(qtop),.b(qnext),.f(irout[4:0]),.s(aluout));
 dpram #(18,12,4096) dpram0(.clk(clk),.load1(dbus2ram),.addr1(abus),.addr2(pcnext),.d1(dbus),.q1(ramout),.q2(irout));

 always @ (pcinc or jump or cs or pcout or irout or poppc or qtop)
   if(cs == `IDLEB) pcnext = pcout;
   else if(pcinc)  pcnext = pcout + 1;
   else if(jump)  pcnext = irout[11:0];
	else if(poppc) pcnext = qtop[11:0]; // added to support POPPC opcode
   else pcnext = 12'hxxx;

 always @ (posedge clk or negedge reset)
   if(!reset) pcout <= 0;
   else if(pcinc || jump || poppc) pcout <= pcnext; // mod to support poppc

 // update the output ports
 always @ (posedge clk)
   if(dbus2obuf) out[irout[2:0]] <= dbus ;
	
 // the 8 ports are external
 assign out0 = out[0];
 assign out1 = out[1];
 assign out2 = out[2];
 assign out3 = out[3];
 assign out4 = out[4];
 assign out5 = out[5];
 assign out6 = out[6];
 assign out7 = out[7];
 
 // the 8 ports are external
 assign in[0] = in0;
 assign in[1] = in1;
 assign in[2] = in2;
 assign in[3] = in3;
 assign in[4] = in4;
 assign in[5] = in5;
 assign in[6] = in6;
 assign in[7] = in7;

 always @ (posedge clk or negedge reset)
   if(!reset) qthru <= 0;
   else qthru <= thru;

 always @ (ir2abus or qtop2abus or qnext2abus or irout or qtop or qnext)
 if(ir2abus) abus = irout[11:0];
 else if(qtop2abus) abus = qtop[11:0];
 else if(qnext2abus) abus = qnext[11:0];
 else abus = 12'hxxx;

 assign dbus = ir2dbus ? {{6{irout[11]}},irout[11:0]} : 18'hzzzz;
 assign dbus = qtop2dbus ? qtop : 18'hzzzz;
 assign dbus = alu2dbus ? aluout : 18'hzzzz;
 assign dbus = ram2dbus ? ramout : 18'hzzzz;
 assign dbus = in2dbus ? in[irout[2:0]] : 18'hzzzz;
 assign dbus = pushpc ? pcnext : 18'hzzzz; // added by BRL4 to support PUSHPC opcode

 always @(cs or irout or qtop)
   begin
     halt = 0; pcinc = 0; jump = 0; push = 0; pop = 0; pop2 = 0; 
	  thru = 0; dbus2qtop = 0; dbus2ram = 0; dbus2obuf = 0; ir2dbus = 0; 
	  qtop2dbus = 0; alu2dbus = 0; ram2dbus = 0; in2dbus = 0; ir2abus = 0; 
	  qtop2abus = 0; qnext2abus = 0; 
	  poppc = 0; pushpc = 0; // added to support poppc and pushpc
	  
  if(cs == `EXEC)
    case(irout[15:12])
      `PUSHI:
         begin
           ir2dbus = 1; dbus2qtop = 1; push = 1; pcinc = 1;
         end	 
      `PUSH: 
         begin
           ir2abus = 1; push = 1; thru = 1; pcinc = 1;
         end
      `POP:
         begin
           ir2abus = 1; qtop2dbus = 1; dbus2ram = 1; pop = 1; pcinc = 1;
         end
      `JMP: jump = 1;
      `JZ:
           begin
             if(qtop == 0) jump = 1;
             else pcinc = 1;
             pop = 1;
           end
      `JNZ: 
          begin
            if(qtop != 0) jump = 1;
             else pcinc = 1;
             pop = 1;
          end
      `LD:
         begin
           qtop2abus = 1; thru = 1; pcinc = 1;
         end
      `ST:
         begin
           qnext2abus = 1; qtop2dbus = 1; dbus2ram = 1; pop2 = 1; pcinc = 1;
         end
		// added by BRL4 //////////////////////////////// 
		  `PPPC:
		  if(irout[0] == 0) //pushpc
			begin
				pushpc = 1; dbus2qtop = 1; push = 1; pcinc = 1;
			end
			
			else //poppc
			begin
				poppc = 1;  pop = 1;
			end
		// end addition ////////////////////////////////
     `IN:
         begin
          in2dbus = 1; dbus2qtop = 1; push = 1; pcinc = 1;
         end
      `OUT:
         begin
           qtop2dbus = 1; dbus2obuf = 1; pop = 1; pcinc = 1;
         end
      `OP:
           begin
             alu2dbus = 1; dbus2qtop = 1; pcinc = 1;
             if(irout[4] == 0) pop = 1; //binary operations
				 // do nothing for bnot, neg, not
				 if(irout[4:0]==5'b10011) push = 1;  // dup
				 if(irout[4:0]==5'b10111) pop = 1;  // drop
				 if(irout[4:0]==5'b11111) push = 1; // over
				 if(irout[4:0]==5'b11011) pop = 1;  // dropnext
				 if(irout[4:0]==5'b10100) pop = 1;  // arithmetic shift right
				 if(irout[4:0]==5'b10101) pop = 1;  //fixed mult
           end
			  
       default:
         begin
           pcinc = 1 ; // halt=1; change opcode zero to NOP
         end
      endcase
  end

endmodule


//========== KEYBOARD TO ASCII LUT ==========//

// lookup table for keyboard key
// given a scancode, output an 8-bit ascii char

/*
    NOTE: characters not handled in this LUT
    - shift         (handled outside LUT)
    - caps lock     (handled outside LUT)
    - ctrl          (not handled)
    - Windows key   (not handled)
    - alt           (not handled)
    - menu key      (not handled)
    
    TODO: the ones handled outside the LUT should prob be handled in the always block
*/
module keyLUT(scancode, shift, caps_lock, ascii);
  input  [7:0] scancode;
  input        shift;
  input        caps_lock;
  output [7:0] ascii;

  assign ascii =
        //first row (` to <backspace>)
        (scancode == 8'd144 & ~shift) ? 8'h60: //` 
        (scancode == 8'd144 & shift)  ? 8'h7E: //~
        (scancode == 8'd145 & ~shift) ? 8'h31: //1
        (scancode == 8'd145 & shift)  ? 8'h21: //!
        (scancode == 8'd146 & ~shift) ? 8'h32: //2
        (scancode == 8'd146 & shift)  ? 8'h40: //@
        (scancode == 8'd147 & ~shift) ? 8'h33: //3
        (scancode == 8'd147 & shift)  ? 8'h23: //#
        (scancode == 8'd148 & ~shift) ? 8'h34: //4
        (scancode == 8'd148 & shift)  ? 8'h24: //$
        (scancode == 8'd149 & ~shift) ? 8'h35: //5
        (scancode == 8'd149 & shift)  ? 8'h25: //%
        (scancode == 8'd150 & ~shift) ? 8'h36: //6
        (scancode == 8'd150 & shift)  ? 8'h5E: //^
        (scancode == 8'd151 & ~shift) ? 8'h37: //7
        (scancode == 8'd151 & shift)  ? 8'h26: //&
        (scancode == 8'd152 & ~shift) ? 8'h38: //8
        (scancode == 8'd152 & shift)  ? 8'h2A: //*
        (scancode == 8'd153 & ~shift) ? 8'h39: //9
        (scancode == 8'd153 & shift)  ? 8'h28: //(
        (scancode == 8'd154 & ~shift) ? 8'h30: //0
        (scancode == 8'd154 & shift)  ? 8'h29: //)
        (scancode == 8'd155 & ~shift) ? 8'h2D: //-
        (scancode == 8'd155 & shift)  ? 8'h5F: //_
        (scancode == 8'd156 & ~shift) ? 8'h3D: //=
        (scancode == 8'd156 & shift)  ? 8'h2B: //+
        (scancode == 8'd157)          ? 8'h08: //<backspace> TODO: ***
        
        //second row (<tab> to \)
        (scancode == 8'd158)          ? 8'h09: //<tab> TODO: ***
        (scancode == 8'd159 & ~(shift | caps_lock)) ? 8'h71: //q
        (scancode == 8'd159 &  (shift | caps_lock)) ? 8'h51: //Q
        (scancode == 8'd160 & ~(shift | caps_lock)) ? 8'h77: //w
        (scancode == 8'd160 &  (shift | caps_lock)) ? 8'h57: //W
        (scancode == 8'd161 & ~(shift | caps_lock)) ? 8'h65: //e
        (scancode == 8'd161 &  (shift | caps_lock)) ? 8'h45: //E
        (scancode == 8'd162 & ~(shift | caps_lock)) ? 8'h72: //r
        (scancode == 8'd162 &  (shift | caps_lock)) ? 8'h52: //R
        (scancode == 8'd163 & ~(shift | caps_lock)) ? 8'h74: //t
        (scancode == 8'd163 &  (shift | caps_lock)) ? 8'h54: //T
        (scancode == 8'd164 & ~(shift | caps_lock)) ? 8'h79: //y
        (scancode == 8'd164 &  (shift | caps_lock)) ? 8'h59: //Y
        (scancode == 8'd165 & ~(shift | caps_lock)) ? 8'h75: //u
        (scancode == 8'd165 &  (shift | caps_lock)) ? 8'h55: //U
        (scancode == 8'd166 & ~(shift | caps_lock)) ? 8'h69: //i
        (scancode == 8'd166 &  (shift | caps_lock)) ? 8'h49: //I
        (scancode == 8'd167 & ~(shift | caps_lock)) ? 8'h6F: //o
        (scancode == 8'd167 &  (shift | caps_lock)) ? 8'h4F: //O
        (scancode == 8'd168 & ~(shift | caps_lock)) ? 8'h70: //p
        (scancode == 8'd168 &  (shift | caps_lock)) ? 8'h50: //P
        (scancode == 8'd169 & ~shift) ? 8'h5B: //[
        (scancode == 8'd169 &  shift) ? 8'h7B: //{
        (scancode == 8'd170 & ~shift) ? 8'h5D: //]
        (scancode == 8'd170 &  shift) ? 8'h7D: //}
        (scancode == 8'd171 & ~shift) ? 8'h5C: //\
        (scancode == 8'd171 &  shift) ? 8'h7C: //|
        
        //third row (<caps lock> to <enter>)
        //(scancode == 8'd172) caps lock handled outside the LUT
        (scancode == 8'd173 & ~(shift | caps_lock)) ? 8'h61: //a
        (scancode == 8'd173 &  (shift | caps_lock)) ? 8'h41: //A
        (scancode == 8'd174 & ~(shift | caps_lock)) ? 8'h73: //s
        (scancode == 8'd174 &  (shift | caps_lock)) ? 8'h53: //S
        (scancode == 8'd175 & ~(shift | caps_lock)) ? 8'h64: //d
        (scancode == 8'd175 &  (shift | caps_lock)) ? 8'h44: //D
        (scancode == 8'd176 & ~(shift | caps_lock)) ? 8'h66: //f
        (scancode == 8'd176 &  (shift | caps_lock)) ? 8'h46: //F
        (scancode == 8'd177 & ~(shift | caps_lock)) ? 8'h67: //g
        (scancode == 8'd177 &  (shift | caps_lock)) ? 8'h47: //G
        (scancode == 8'd178 & ~(shift | caps_lock)) ? 8'h68: //h
        (scancode == 8'd178 &  (shift | caps_lock)) ? 8'h48: //H
        (scancode == 8'd179 & ~(shift | caps_lock)) ? 8'h6A: //j
        (scancode == 8'd179 &  (shift | caps_lock)) ? 8'h4A: //J
        (scancode == 8'd180 & ~(shift | caps_lock)) ? 8'h6B: //k
        (scancode == 8'd180 &  (shift | caps_lock)) ? 8'h4B: //K
        (scancode == 8'd181 & ~(shift | caps_lock)) ? 8'h6C: //l
        (scancode == 8'd181 &  (shift | caps_lock)) ? 8'h4C: //L
        (scancode == 8'd182 & ~shift) ? 8'h3B: //;
        (scancode == 8'd182 &  shift) ? 8'h3A: //:
        (scancode == 8'd183 & ~shift) ? 8'h27: //'
        (scancode == 8'd183 &  shift) ? 8'h22: //"
        (scancode == 8'd184)          ? 8'h0A: //<enter> TODO: ***

        //fourth row (z to /)
        //(scancode == 8'd185) left-shift handled outside the LUT
        (scancode == 8'd186 & ~(shift | caps_lock))  ? 8'h7A: //z
        (scancode == 8'd186 &  (shift | caps_lock))  ? 8'h5A: //Z
        (scancode == 8'd187 & ~(shift | caps_lock))  ? 8'h78: //x
        (scancode == 8'd187 &  (shift | caps_lock))  ? 8'h58: //X
        (scancode == 8'd188 & ~(shift | caps_lock))  ? 8'h63: //c
        (scancode == 8'd188 &  (shift | caps_lock))  ? 8'h43: //C
        (scancode == 8'd189 & ~(shift | caps_lock))  ? 8'h76: //v
        (scancode == 8'd189 &  (shift | caps_lock))  ? 8'h56: //V
        (scancode == 8'd190 & ~(shift | caps_lock))  ? 8'h62: //b
        (scancode == 8'd190 &  (shift | caps_lock))  ? 8'h42: //B
        (scancode == 8'd191 & ~(shift | caps_lock))  ? 8'h6E: //n
        (scancode == 8'd191 &  (shift | caps_lock))  ? 8'h4E: //N
        (scancode == 8'd192 & ~(shift | caps_lock))  ? 8'h6D: //m
        (scancode == 8'd192 &  (shift | caps_lock))  ? 8'h4D: //M
        (scancode == 8'd193 & ~shift)  ? 8'h2C: //,
        (scancode == 8'd193 &  shift)  ? 8'h3C: //<
        (scancode == 8'd194 & ~shift)  ? 8'h2E: //.
        (scancode == 8'd194 &  shift)  ? 8'h3E: //>
        (scancode == 8'd195 & ~shift)  ? 8'h2F: ///
        (scancode == 8'd195 &  shift)  ? 8'h3F: //?
        //(scancode == 8'd196) right-shift handled outside the LUT
        
        //last row (space)
        (scancode == 8'd200)           ? 8'h20: //<space>
						
        8'h00;   // default NULL
endmodule


////////////////// end file //////////////////