// Implements:
// - SD Card controller and the SD Card SPI Interface.
// - The SD Card is used to simulate 4 X1215 disk packs for 2 X1215 disk Units
//
// Name   : sd_card3.v
// Version: 3
// Date   : 13-10-2010
// Author : Theo Engel
// Contact: Info@theoengel.nl 


module SD_Card( // Controller Interface
                iCmd, iScAd, oCmd_Ack, oOperational, oRW_Ready,
                // Host data interface (sdcard=>host)
                oBreak, oRsD_DATA, // byte from sdcard
                // Host data interface (host=> sdcard)
                iWsD_DATA, oAckOutput,
				// SD Card Interface
				SD_CMD, SD_DAT, SD_DAT3, SD_CLK, Response,
				// reset and clock
				iRST_n, iCLK
                );

// Controller Interface
input  wire [1:0]  iCmd;     	// Read (01), Write (10), or None (00) (default)
input  wire [31:0] iScAd;    	// Byte Address
output wire        oCmd_Ack; 	// signal host that read/write command is accepted

output wire        oBreak;		// signal an input break (byte available from sdcard)
output wire [7:0]  oRsD_DATA;	// input data byte from sdcard (if oBreak == 1)

output wire        oAckOutput;	// signal ack to inform the host that output byte is being send to sd card
input  wire [7:0]  iWsD_DATA;	// output data byte to sdcard 

output wire        oRW_Ready;	// signal host that read/write command is processed
output wire        oOperational;// sd card is in init state/ready
output wire [7:0]  Response;    // sd card response byte of a write command

//	SD Card Interface
input  wire        SD_DAT;		// SD Card Data (SPI: DO output on card)
output wire        SD_DAT3;		// SD Card Data 3 (SPI: CS, chip select input on card)
output wire        SD_CMD;		// SD Card Command Signal (SPI: DI input on card)
output wire        SD_CLK;		// SD Card Clock (SPI: Clock input on card)

//  reset and clock
input  wire iRST_n;
input  wire iCLK;

// Internal wiring
wire [7:0]  mTsD_DATA;
wire        mTsD_Done;
wire        mTsD_Start;
wire        mRsD_Ready;
wire        mSD_Select;


// Sd Card Controller
SD_Control SD_C(.iCmd(iCmd),.iAddr(iScAd),.oCmd_Ack(oCmd_Ack),.oOperational(oOperational),
	           
	           .oRW_Ready(oRW_Ready),
	           
	           .oSD_Select_n(mSD_Select),
	           
	           .iWsD_DATA(iWsD_DATA),.oAckOutData(oAckOutput),
	           .oTxD_DATA(mTsD_DATA),.iTxD_Done(mTsD_Done),.oTxD_Start(mTsD_Start),  // byte to sdcard spi interface
	           .oBreak(oBreak), .iRxD_DATA(oRsD_DATA),.iRxD_Ready(mRsD_Ready), 		 // byte from sdcard spi interface
	           .Response(Response), .iRST_n(iRST_n), .iCLK(iCLK)
               );

// Sd Card SPI Interface
SD_SPI	  SD_I(	//	HOST
				.iTxD_DATA(mTsD_DATA),.oTxD_Done(mTsD_Done),.iTxD_Start(mTsD_Start), // byte to sdcard
				
				.oRxD_DATA(oRsD_DATA),.oRxD_Ready(mRsD_Ready), 						 // byte from sdcard
				
				.iSD_Select_n(mSD_Select), .init(oOperational),
				.iRST_n(iRST_n), .iCLK(iCLK),
				//	SD Card Interface
				.SD_DO(SD_CMD), .SD_DI(SD_DAT), .SD_CS(SD_DAT3), .SD_CK(SD_CLK));

endmodule

///////////////////////////////////////////////////////////////////////////////
// Sd Card Controller

module SD_Control (iCmd, iAddr, oCmd_Ack, oOperational, oRW_Ready, oSD_Select_n,
				   iWsD_DATA, oAckOutData,
				   oTxD_DATA, iTxD_Done, oTxD_Start, 	// byte to sdcard spi interface
				   oBreak, iRxD_DATA, iRxD_Ready, 		// byte from sdcard spi interface
				   Response, iRST_n, iCLK
                   );

input  wire [1:0]  iCmd;          	// 01 == Read Block, 10 == Write Block
input  wire [31:0] iAddr;
output reg         oCmd_Ack;
output reg         oOperational; 	// 1 if Operational (sdcard initialized)
output reg         oRW_Ready;
output wire        oSD_Select_n;
output reg         oAckOutData;
input  wire [7:0]  iWsD_DATA;
output reg  [7:0]  oTxD_DATA;
output reg         oTxD_Start;
input  wire        iTxD_Done;
output reg         oBreak;
input  wire [7:0]  iRxD_DATA;
input  wire        iRxD_Ready;
input  wire        iRST_n;
input  wire        iCLK;
output reg  [7:0]  Response;

reg  [4:0]  SD_state;
wire [47:0] Tx_Mess;
wire        Tx_Mstart;
reg         Tx_Mdone;
reg  [15:0] delay;
reg         idle;
reg     	   init;

assign oSD_Select_n = mSDi_Select_n & mSD_Select_n;

// Message multiplexer
assign Tx_Mstart    = Tx_iMstart | Tx_rwMstart;
assign Tx_Mess      = init ? Tx_rwMess : Tx_iMess;

// Initialize the SD card
// After succesfull initialization, both idle and init are 1'b1 => oOperational is 1'b1
reg                mSDi_Select_n;
reg         [3:0]  SDi_state;
reg                Tx_iMstart;
reg         [47:0] Tx_iMess;

// Send messages and read response R1
always@(posedge iCLK or negedge iRST_n)
begin
	if(!iRST_n)
	begin
		SDi_state     <= 4'b0000;
		Tx_iMstart    <= 1'b0;
		Tx_iMess      <= 48'h000000000000;
		idle          <= 1'b0;
		init          <= 1'b0;
		oOperational  <= 1'b0;
		mSDi_Select_n <= 1'b1;
	end
	else
	begin
		case(SDi_state)
		4'b0000: begin
					mSDi_Select_n <= 1'b1;
					SDi_state     <= 4'b0001;
					idle          <= 1'b0;
					init          <= 1'b0;
					oOperational  <= 1'b0;
				end
		4'b0001: begin
					mSDi_Select_n <= 1'b1;
					SDi_state     <= 4'b0001;
					if(delay == 0)
						SDi_state <= 4'b0010;
			end		
		4'b0010: begin // send reset message
					mSDi_Select_n <= 1'b0;
					Tx_iMess      <= 48'h400000000095; 	// reset message
					Tx_iMstart    <= 1'b1;				// request send
					SDi_state     <= 4'b0011;
				end
		4'b0011: begin // message send ?
					mSDi_Select_n <= 1'b0;
					SDi_state  	  <= 4'b0011;
					Tx_iMstart    <= 1'b0;
					if(Tx_Mdone)
						SDi_state <= 4'b0100;
				end
		4'b0100: begin // read response
					mSDi_Select_n <= 1'b0;
					if(iRxD_DATA == 8'hff)
						SDi_state <= 4'b0100;
					else
						SDi_state <= 4'b0101;
			end
		4'b0101: begin // is response /01 ?
					mSDi_Select_n <= 1'b1;
					if(iRxD_DATA == 8'h01)
						SDi_state <= 4'b0110;
					else
						SDi_state <= 4'b0001; 			// not /01 => resend reset message
				end
		4'b0110: begin
					mSDi_Select_n <= 1'b1;
					SDi_state     <= 4'b0110;
					if(delay == 0)
						SDi_state <= 4'b0111;
			end
		4'b0111: begin // initialize the sd card
					mSDi_Select_n <= 1'b0;
					Tx_iMess      <= 48'h410000000001; 	// send INIT-CMD
					Tx_iMstart    <= 1'b1;
					SDi_state     <= 4'b1000;
			end
		4'b1000: begin // message send ?
					mSDi_Select_n <= 1'b0;
					SDi_state     <= 4'b1000;
					Tx_iMstart    <= 1'b0;
					if(Tx_Mdone)
						SDi_state <= 4'b1001;
			end
		4'b1001: begin // read response
					mSDi_Select_n <= 1'b0;
					if(iRxD_DATA == 8'hff)
						SDi_state <= 4'b1001;
					else
						SDi_state <= 4'b1010;
				end
		4'b1010: begin
					SDi_state     <= 4'b1010;
					mSDi_Select_n <= 1'b1;
					if((iRxD_DATA == 8'h00) && idle )
					begin
						init      <= 1'b1;
						SDi_state <= 4'b1011;
					end
					else
					begin
						if((iRxD_DATA == 8'h01))
						begin
							idle      <= 1'b1;
						end
						SDi_state <= 4'b0110; 			// resend INIT command
					end
				end
		4'b1011: begin // initialized; stop
					mSDi_Select_n <= 1'b1;
					if(idle & init)
					begin
						oOperational  <= 1'b1;			// set card operational
						SDi_state     <= 4'b1011;
					end
					else
						SDi_state <= 4'b0000;
			end
		default: begin
					SDi_state     <= 4'b0000;
					Tx_iMstart    <= 1'b0;
					idle          <= 1'b0;
					init          <= 1'b0;
					oOperational  <= 1'b0;
					mSDi_Select_n <= 1'b1;
				end
		endcase
	end
end

		
// Read/Write a Sector from/to the SDcard
reg        mSD_Select_n;
reg        Tx_rwMstart;
reg [47:0] Tx_rwMess;
reg [9:0]  count;
reg        Tx_Sector;
reg [7:0]  oByte;

always@(posedge iCLK or negedge iRST_n)
begin
	if(!iRST_n)
	begin
		SD_state      <= 5'b00000;
		Tx_rwMstart   <= 1'b0;
		Tx_rwMess     <= 48'h000000000000;
		mSD_Select_n  <= 1'b1;
		count         <= 10'b0000000000;
		oCmd_Ack      <= 1'b0;
		oRW_Ready     <= 1'b0;
		oBreak        <= 1'b0;
		Tx_Sector     <= 1'b0;
		oByte         <= 8'h00;
		oAckOutData   <= 1'b0;
		Response      <= 8'h00;
	end
	else
	begin
		case(SD_state)				
		5'b00000: begin
					SD_state 		<= 5'b00000;
					Tx_rwMstart     <= 1'b0;
					Tx_rwMess       <= 48'h000000000000;				
					mSD_Select_n 	<= 1'b1;
					count           <= 10'b0000000000;
					oCmd_Ack  		<= 1'b0;
					oRW_Ready 		<= 1'b0;
					oBreak 			<= 1'b0;
					Tx_Sector       <= 1'b0;
					oByte           <= 8'h00;
					oAckOutData     <= 1'b0;
					Response        <= 8'h00;				
					if(oOperational)
						SD_state <= 5'b00001;
				end
		5'b00001: begin
					mSD_Select_n 	<= 1'b1;
					SD_state 		<= 5'b00001;
					if(delay == 0)
						SD_state 	<= 5'b00010;
				end
		5'b00010: begin
					mSD_Select_n 	<= 1'b1;
					count           <= 10'b0000000000;	// block length (512 + 2 CRC bytes) 
					SD_state 		<= 5'b00010;				
					if(iCmd == 2'b01) 					// read a block
					begin
						oCmd_Ack  <= 1'b1;
						SD_state  <= 5'b00011;
					end
					if(iCmd == 2'b10) 					// write a block
					begin
						oCmd_Ack  <= 1'b1;
						SD_state  <= 5'b01010;
					end
				end
		/////// Input
		5'b00011: begin // send command read a block (cmd 0x11)
					oCmd_Ack  		<= 1'b0;
					Tx_rwMess   	<= {8'h51, iAddr, 8'hFF}; //sector 0 disk 1 == /2000 + /280)
					Tx_rwMstart 	<= 1'b1;
					mSD_Select_n 	<= 1'b0;
					SD_state 		<= 5'b00100;
				end
		5'b00100: begin
					Tx_rwMstart 	<= 1'b0;
					SD_state  		<= 5'b00100;
					if(Tx_Mdone)
						SD_state <= 5'b00101;
				end
		5'b00101: begin
					SD_state  		<= 5'b00101;			
					if(iRxD_Ready && (iRxD_DATA == 8'h00)) 	// R1 response == 0
						SD_state 	<= 5'b00110;
				end
		5'b00110: begin
					SD_state  		<= 5'b00110;
					if(iRxD_Ready && (iRxD_DATA == 8'hFE))
						SD_state    <= 5'b00111;			// Data token
				end
		5'b00111: begin  // read block (514 bytes: 512 data bytes + 2 crc bytes)
					SD_state 		<= 5'b00111;
					if(iRxD_Ready)
					begin
						count 	 	<= count + 1'b1;
						oBreak 	 	<= 1'b1;
						SD_state 	<= 5'b01000;
					end
				end
		5'b01000: begin
					oBreak 			<= 1'b0;
					SD_state 		<= 5'b00111;
					if(count == 514)
					begin
						mSD_Select_n <= 1'b1;
						oRW_Ready    <= 1'b1;
						SD_state     <= 5'b01001;
					end
				end			
		5'b01001: begin
					mSD_Select_n 	<= 1'b1;
					oRW_Ready 		<= 1'b1;
					SD_state 		<= 5'b00000;
				end
		//////// Output	
		5'b01010: begin // send command write block (cmd 0x18)
					oCmd_Ack  		<= 1'b0;
					Tx_rwMess   	<= {8'h58, iAddr, 8'hFF}; 
					Tx_rwMstart 	<= 1'b1;
					mSD_Select_n 	<= 1'b0;
					Tx_Sector       <= 1'b1;	// set write block bit
					SD_state 		<= 4'b1011;	
				end
		5'b01011: begin
					Tx_rwMstart 	<= 1'b0;
					SD_state  		<= 5'b01011;
					if(Tx_Mdone)
						SD_state <= 5'b01100;	// message send, R1 response being read
				end				
		5'b01100: begin
					SD_state  		<= 5'b01100;
					oByte           <= 8'hff;	// 1st FF								
					if(iRxD_Ready && (iRxD_DATA == 8'h00)) // R1 response == 0
						SD_state <= 5'b01101;
				end
				// R1 response is OK; send FF FF FE <block>
		5'b01101: begin
					SD_state  		<= 5'b01101;
					oByte           <= 8'hff;	// 1st FF
					if(iTxD_Done)
						SD_state  <= 5'b01110;
				end
		5'b01110: begin
					SD_state  		<= 5'b01110;	
					oByte           <= 8'hff;	// 2nd FF
					if(iTxD_Done)
						SD_state  <= 5'b01111;
				end
		5'b01111: begin
					SD_state  		<= 5'b01111;
					oByte           <= 8'hfe;
					if(iTxD_Done)
						SD_state  <= 5'b10000;
				end					
		5'b10000: begin
					// request a new byte iWsD_DATA while transfering a byte oByte to the sdcard
					SD_state  		<= 5'b10001;	
					oByte           <= iWsD_DATA;	// output data byte to the sdcard
					oAckOutData     <= 1'b1;        // iWsD_DATA can be reloaded
				end
		5'b10001: begin
					SD_state  		<= 5'b10001;
					oAckOutData     <= 1'b0;
					if(iTxD_Done)
					begin
						count <= count + 1'b1;
						SD_state  <= 5'b10011;
					end
				end
		5'b10011: begin
					if(count == 511) //all bytes in sector after this one are 0xFF
					begin
						oByte       <= 8'hff;
						SD_state    <= 5'b10100;
					end
					else
						SD_state  <= 5'b10000;
				end
		5'b10100: begin
					SD_state  		<= 5'b10100;	
					if(iRxD_Ready && (iRxD_DATA != 8'hff))
					begin
						Response    <= iRxD_DATA;
						SD_state 	<= 5'b10101;
					end							
				end
		5'b10101: begin //busy
					SD_state  		<= 5'b10101;	
					if(iRxD_Ready && (iRxD_DATA != 8'h00))
						SD_state 	<= 5'b01001;
				end				
		default: begin
					SD_state 		<= 5'b00000;
					Tx_rwMstart     <= 1'b0;
					Tx_rwMess       <= 48'h000000000000;				
					mSD_Select_n 	<= 1'b1;
					count           <= 10'b0000000000;
					oCmd_Ack  		<= 1'b0;
					oRW_Ready 		<= 1'b0;
					oBreak 			<= 1'b0;
					Tx_Sector       <= 1'b0;
					oByte           <= 8'h00;
					oAckOutData     <= 1'b0;
					Response        <= 8'h00;
				end
		endcase
	end
end

// Deselect (if oSD_Select_n becomes 1) resets the delay counter
reg PreSD_Select;
always @ (posedge iCLK)
begin
	if({PreSD_Select,oSD_Select_n} == 2'b01)
		delay <= 16'h0000;
	else
		delay <= delay + 1'b1;
	PreSD_Select <= oSD_Select_n;
end

// send a 48 bit message and read the response R1 into iRxD_DATA,
// or, (if Tx_Sector == 1) send a 48 bit message, read the response R1 into iRxD_DATA and output a sector.
reg [3:0] Tx_St;
reg [3:0] Tx_St_next;

always @(*)
begin
	Tx_St_next = Tx_St;
		case(Tx_St)
	4'b0000: begin
			if(Tx_Mstart)
				begin
					oTxD_Start   = 1'b0;
					oTxD_DATA    = 8'h00;
					Tx_Mdone     = 1'b0;
					Tx_St_next   = 4'b0001;
				end
			else
				begin
					oTxD_Start   = 1'b0;
					oTxD_DATA    = 8'h00;
					Tx_Mdone     = 1'b0;
					Tx_St_next   = 4'b0000;			
				end
			end
	4'b0001: begin
				oTxD_DATA    = 8'hff;
				oTxD_Start   = 1'b1;
				Tx_Mdone     = 1'b0;
				Tx_St_next   = 4'b0001;
				if(iTxD_Done)
				begin
					oTxD_Start = 1'b0;
					Tx_St_next = 4'b0010;
				end
			end
	4'b0010: begin
				oTxD_DATA    = 8'hff;
				oTxD_Start   = 1'b0;
				Tx_Mdone     = 1'b0;
				Tx_St_next   = 4'b0010;
				if(iRxD_Ready)
				begin
					if(iRxD_DATA == 8'hff)
						Tx_St_next   = 4'b0011;
					else
						Tx_St_next   = 4'b0001;
				end
			end
	4'b0011: begin
				oTxD_DATA    = Tx_Mess[47:40];
				oTxD_Start   = 1'b1;
				Tx_Mdone     = 1'b0;
				Tx_St_next   = 4'b0011;
				if(iTxD_Done)
				begin
					oTxD_Start = 1'b0;
					Tx_St_next = 4'b0100;
				end
			end
	4'b0100: begin
				oTxD_DATA    = Tx_Mess[39:32];
				oTxD_Start   = 1'b1;
				Tx_Mdone     = 1'b0;
				Tx_St_next   = 4'b0100;
				if(iTxD_Done)
				begin
					oTxD_Start = 1'b0;
					Tx_St_next = 4'b0101;
				end
			end
	4'b0101: begin
				oTxD_DATA  	 = Tx_Mess[31:24];
				oTxD_Start   = 1'b1;
				Tx_Mdone     = 1'b0;
				Tx_St_next   = 4'b0101;
				if(iTxD_Done)
				begin
					oTxD_Start = 1'b0;
					Tx_St_next = 4'b0110;
				end
			end	
	4'b0110: begin
				oTxD_DATA    = Tx_Mess[23:16];
				oTxD_Start   = 1'b1;
				Tx_Mdone     = 1'b0;
				Tx_St_next   = 4'b0110;			
				if(iTxD_Done)
				begin
					oTxD_Start = 1'b0;
					Tx_St_next = 4'b0111;
				end
			end				
	4'b0111: begin
				oTxD_DATA    = Tx_Mess[15:8];
				oTxD_Start   = 1'b1;
				Tx_Mdone     = 1'b0;
				Tx_St_next   = 4'b0111;	
				if(iTxD_Done)
				begin
					oTxD_Start = 1'b0;
					Tx_St_next = 4'b1000;
				end
			end						
	4'b1000: begin
				oTxD_DATA    = Tx_Mess[7:0];
				oTxD_Start   = 1'b1;
				Tx_Mdone     = 1'b0;
				Tx_St_next   = 4'b1000;	
				if(iTxD_Done)
				begin
					oTxD_Start = 1'b0;
					Tx_St_next = 4'b1001;
				end
			end					
	4'b1001: begin
				oTxD_DATA    = 8'hff;
				oTxD_Start   = 1'b1;
				Tx_Mdone     = 1'b1;		// signal message sent and response being read
				Tx_St_next   = 4'b1001;
				if(iTxD_Done || iRxD_Ready)
				begin
					if(iRxD_Ready && (iRxD_DATA != 8'hff))
						Tx_St_next = 4'b1010;
					else if(iTxD_Done)
							oTxD_Start = 1'b0;
				end
			end
	4'b1010: begin
				oTxD_DATA    = 8'hff;
				Tx_Mdone     = 1'b0;
				oTxD_Start   = 1'b0;
				if(Tx_Sector)
					Tx_St_next   = 4'b1011;	// output sector to sd card
				else
					Tx_St_next   = 4'b0000; // ready with just sending a message
			end
	4'b1011: begin // output a disk sector to the sdcard
				oTxD_DATA    = oByte;
				Tx_Mdone     = 1'b0;
				oTxD_Start   = 1'b1;
				Tx_St_next   = 4'b1011;
				if(iTxD_Done)
				begin
					oTxD_Start = 1'b0;
					Tx_St_next = 4'b1100;
				end
				if(oRW_Ready)
				begin
					oTxD_Start = 1'b0;
					Tx_St_next = 4'b0000;
				end		
			end
	4'b1100: begin
				oTxD_DATA    = oByte;
				Tx_Mdone     = 1'b0;
				oTxD_Start   = 1'b0;
				if(oRW_Ready)
					Tx_St_next = 4'b0000;
				else
					Tx_St_next = 4'b1011;
			end					
	default: begin
				Tx_St_next   = 4'b0000;
				oTxD_Start   = 1'b0;
				Tx_Mdone     = 1'b0;
				oTxD_DATA    = 8'h00;
			end
		endcase
end	

always @(posedge iCLK or negedge iRST_n)
begin
	if(! iRST_n) Tx_St <= 4'b0000; else Tx_St <= Tx_St_next;
end

endmodule

///////////////////////////////////////////////////////////////////////////////
// SD Card SPI interface

module SD_SPI(	//	HOST
				iTxD_DATA,oTxD_Done,iTxD_Start, // to sdcard
				oRxD_DATA,oRxD_Ready,           // from sdcard
				iSD_Select_n, init,
				iRST_n,iCLK,
				//	SD Card
				SD_DO, SD_DI, SD_CS, SD_CK);
// Host interface
input wire  [7:0] iTxD_DATA; 	// data to sdcard
input wire  iTxD_Start;
output reg  oTxD_Done;

output reg  [7:0] oRxD_DATA; 	// data from sdcrad
output reg  oRxD_Ready;

input wire  iSD_Select_n;   	// 0 => card selected;
input wire  init;				// init == 1 if card is in init state (controls clock speed)

input wire  iRST_n, iCLK;

// SD Card interface
input  wire SD_DI; 
output wire SD_CS;
output reg  SD_DO;
output reg  SD_CK;
//

// Clock generator SD_CK for the SD Card
// Clock Divide
//  7   6   5  4  3  2 1 0
// 256 128 64 32 16  8 4 2
//
reg  [7:0] clk_divide;              // Clock divide counter   
wire       clk_gen;
wire [2:0] divide_factor; 			// by 50 Mhz by 128 * 2 (192 khz)
/* clock divider */
assign     divide_factor = init ? 3'b110 : 3'b110;
assign     clk_gen       = clk_divide[divide_factor];

always @ (posedge iCLK)
begin
	clk_divide <= clk_divide + 8'h01;     
end

/* Generating the SD Card clock */
always @ (posedge clk_gen or negedge iRST_n) 
begin
	if(! iRST_n)
        SD_CK <= 1'b0;
    else
		if (iSD_Select_n)
			SD_CK <= 1'b0; 		 
		else 
			SD_CK <= ~SD_CK;
end

assign SD_CS = iSD_Select_n;

reg  [7:0] filter;
wire [7:0] filter_next;
reg        sdclk;
wire       sdclk_next;
wire       fall_edge;
wire       rise_edge;
reg  [7:0] rDATA;       // read data byte
reg  [2:0] Cntw;		// write bit counter
reg  [2:0] Cntr;        // read bit counter
reg        rState;      // read state

always@(posedge iCLK or negedge iRST_n)
if(!iRST_n)
	begin
		filter <= 8'h00;
		sdclk  <= 1'b0;
	end
else
	begin
		filter <= filter_next;
		sdclk  <= sdclk_next;
	end

assign filter_next = {SD_CK, filter[7:1]};
assign sdclk_next  = (filter == 8'hff) ? 1'b1 :
					 (filter == 8'h00) ? 1'b0 :	 sdclk;

assign fall_edge = sdclk  & ~sdclk_next;
assign rise_edge = ~sdclk & sdclk_next;

always@(posedge iCLK or posedge SD_CS or negedge iRST_n)
begin
	if(!iRST_n | SD_CS)
	begin
		Cntw       <= 3'b111;
        Cntr       <= 3'b000;
        SD_DO      <= 1'b1;
		oTxD_Done  <= 1'b0;
		rDATA      <= 8'hff;
		oRxD_DATA  <= 8'hff;
		oRxD_Ready <= 1'b0;
		rState     <= 1'b0;		
    end
    else
    begin
		oTxD_Done  <= 1'b0;
		oRxD_Ready <= 1'b0;	
		if(fall_edge) //write on negedge
		begin
			if(iTxD_Start)
			begin
				Cntw   <= Cntw - 3'b001;						
				SD_DO  <= iTxD_DATA[Cntw];
				if(Cntw == 3'b000)
					oTxD_Done <= 1'b1;
			end	
			else
			begin
				Cntw  <= 3'b111;
				SD_DO <= 1'b1;
			end
		end        
        if(rise_edge) //read on posedge
		begin
			case(rState)
			1'b0: begin
					rState     <= 1'b0;
					oRxD_DATA  <= 8'hff;
					if(~ SD_DI) // sync with start bit
					begin
						Cntr   <= 3'b001;
						rDATA  <= 8'hfe;
						rState <= 1'b1;
					end
					else
					begin
						Cntr   <= Cntr + 3'b001;
						if(Cntr == 3'b111)
						begin
							oRxD_Ready <= 1'b1;
						end
					end
				end
			1'b1: begin
					rState      <= 1'b1;
					Cntr        <= Cntr + 3'b001;
					rDATA       <= {rDATA[6:0],SD_DI};
					if(Cntr == 3'b111)
					begin
						oRxD_DATA  <= {rDATA[6:0],SD_DI};
						oRxD_Ready <= 1'b1;
					end
				end
			endcase				
		end
	end
end
endmodule
